@@ -8,12 +8,15 @@ import pytest
 
 import cupy
 from cupy import cuda
+from cupy.cuda import runtime
 from cupy.random import _generator
 from cupy import testing
 from cupy.testing import _attr
 from cupy.testing import _condition
 from cupy.testing import _hypothesis
 
+from cupy_tests.random_tests import common_distributions
+
 
 def numpy_cupy_equal_continuous_distribution(significance_level, name='xp'):
     """Decorator that tests the distributions of NumPy samples and CuPy ones.
@@ -40,7 +43,7 @@ def numpy_cupy_equal_continuous_distribution(significance_level, name='xp'):
             assert cupy_result is not None
             assert numpy_result is not None
             d_plus, d_minus, p_value = \
-                two_sample_Kolmogorov_Smirnov_test(
+                common_distributions.two_sample_Kolmogorov_Smirnov_test(
                     cupy.asnumpy(cupy_result), numpy_result)
             if p_value < significance_level:
                 message = '''Rejected null hypothesis:
@@ -52,123 +55,23 @@ D- (cupy > numpy): %f''' % (p_value, d_plus, d_minus)
     return decorator
 
 
-def two_sample_Kolmogorov_Smirnov_test(observed1, observed2):
-    """Computes the Kolmogorov-Smirnov statistic on 2 samples
-
-    Unlike `scipy.stats.ks_2samp`, the returned p-value is not accurate
-    for large p.
-    """
-    assert observed1.dtype == observed2.dtype
-    n1, = observed1.shape
-    n2, = observed2.shape
-    assert n1 >= 100 and n2 >= 100
-    observed = numpy.concatenate([observed1, observed2])
-    indices = numpy.argsort(observed)
-    observed = observed[indices]  # sort
-    ds = numpy.cumsum(numpy.where(indices < n1, -n2, n1).astype(numpy.int64))
-    assert ds[-1] == 0
-    check = numpy.concatenate([observed[:-1] < observed[1:], [True]])
-    ds = ds[check]
-    d_plus = float(ds.max()) / (n1 * n2)
-    d_minus = -float(ds.min()) / (n1 * n2)
-    d = max(d_plus, d_minus)
-    # Approximate p = special.kolmogorov(d * numpy.sqrt(n1 * n2 / (n1 + n2)))
-    p = min(1.0, 2.0 * numpy.exp(-2.0 * d**2 * n1 * n2 / (n1 + n2)))
-    return d_plus, d_minus, p
-
-
 def _get_size(size):
     # CuPy returns an ndarray of shape () even if size=None.
     # cf. NumPy returns a Python scalar if size=None.
     if size is None:
         return ()
-    return cupy.core.get_size(size)
+    return cupy._core.get_size(size)
 
 
-class RandomGeneratorTestCase(unittest.TestCase):
+class RandomGeneratorTestCase(common_distributions.BaseGeneratorTestCase):
 
     target_method = None
 
-    def setUp(self):
-        self.__seed = testing.generate_seed()
-        self.rs = _generator.RandomState(seed=self.__seed)
+    def get_rng(self, xp, seed):
+        return xp.random.RandomState(seed=seed)
 
-    def _get_generator_func(self, *args, **kwargs):
-        assert isinstance(self.target_method, str), (
-            'generate_method must be overridden')
-        f = getattr(self.rs, self.target_method)
-        return lambda: f(*args, **kwargs)
-
-    def _generate_check_repro(self, func, seed):
-        # Sample a random array while checking reproducibility
-        self.rs.seed(seed)
-        x = func()
-        self.rs.seed(seed)
-        y = func()
-        testing.assert_array_equal(
-            x, y,
-            'Randomly generated arrays with the same seed did not match')
-        return x
-
-    def generate(self, *args, **kwargs):
-        # Pick one sample from generator.
-        # Reproducibility is checked by repeating seed-and-sample cycle twice.
-        func = self._get_generator_func(*args, **kwargs)
-        return self._generate_check_repro(func, self.__seed)
-
-    def generate_many(self, *args, **kwargs):
-        # Pick many samples from generator.
-        # Reproducibility is checked only for the first sample,
-        # because it's very slow to set seed every time.
-        _count = kwargs.pop('_count', None)
-        assert _count is not None, '_count is required'
-        func = self._get_generator_func(*args, **kwargs)
-
-        if _count == 0:
-            return []
-
-        vals = [self._generate_check_repro(func, self.__seed)]
-        for _ in range(1, _count):
-            vals.append(func())
-        return vals
-
-    def check_ks(self, significance_level, cupy_len=100, numpy_len=1000):
-        return functools.partial(
-            self._check_ks, significance_level, cupy_len, numpy_len)
-
-    def _check_ks(
-            self, significance_level, cupy_len, numpy_len,
-            *args, **kwargs):
-        assert 'size' in kwargs
-
-        # cupy
-        func = self._get_generator_func(*args, **kwargs)
-        vals_cupy = func()
-        assert vals_cupy.size > 0
-        count = 1 + (cupy_len - 1) // vals_cupy.size
-        vals_cupy = [vals_cupy]
-        for _ in range(1, count):
-            vals_cupy.append(func())
-        vals_cupy = cupy.stack(vals_cupy).ravel()
-
-        # numpy
-        kwargs['size'] = numpy_len
-        dtype = kwargs.pop('dtype', None)
-        numpy_rs = numpy.random.RandomState(self.__seed)
-        vals_numpy = getattr(numpy_rs, self.target_method)(*args, **kwargs)
-        if dtype is not None:
-            vals_numpy = vals_numpy.astype(dtype, copy=False)
-
-        # test
-        d_plus, d_minus, p_value = \
-            two_sample_Kolmogorov_Smirnov_test(
-                cupy.asnumpy(vals_cupy), vals_numpy)
-        if p_value < significance_level:
-            message = '''Rejected null hypothesis:
-p: %f
-D+ (cupy < numpy): %f
-D- (cupy > numpy): %f''' % (p_value, d_plus, d_minus)
-            raise AssertionError(message)
+    def set_rng_seed(self, seed):
+        self.rng.seed(seed)
 
 
 def _xp_random(xp, method_name):
@@ -239,29 +142,30 @@ class TestRandomState(unittest.TestCase):
         ]
 
         for method in methods:
+            if (runtime.is_hip and
+                    method == cupy.cuda.curand.CURAND_RNG_PSEUDO_MT19937):
+                # hipRAND fails for MT19937 with the status code 1000,
+                # HIPRAND_STATUS_NOT_IMPLEMENTED. We use `pytest.raises` here
+                # so that we will be able to find it once hipRAND implement
+                # MT19937 as the imperative `pytest.xfail` immediately rewinds
+                # the control flow and does not run the test.
+                with pytest.raises(KeyError) as e:
+                    rs = cupy.random.RandomState(method=method)
+                assert e.value.args == (1000,)
+                continue
             rs = cupy.random.RandomState(method=method)
             rs.normal()
 
 
-@testing.parameterize(
-    {'a': 1.0, 'b': 3.0},
-    {'a': 3.0, 'b': 3.0},
-    {'a': 3.0, 'b': 1.0},
-)
+@testing.parameterize(*common_distributions.beta_params)
+@testing.with_requires('numpy>=1.17.0')
 @testing.gpu
 @testing.fix_random()
-class TestBeta(RandomGeneratorTestCase):
-
-    target_method = 'beta'
-
-    def test_beta(self):
-        self.generate(a=self.a, b=self.b, size=(3, 2))
-
-    @testing.for_dtypes('fd')
-    @_condition.repeat_with_success_at_least(10, 3)
-    def test_beta_ks(self, dtype):
-        self.check_ks(0.05)(
-            a=self.a, b=self.b, size=2000, dtype=dtype)
+class TestBeta(
+    common_distributions.Beta,
+    RandomGeneratorTestCase
+):
+    pass
 
 
 @testing.parameterize(
@@ -322,25 +226,14 @@ class TestDirichlet(RandomGeneratorTestCase):
     # TODO(kataoka): add distribution test
 
 
-@testing.parameterize(
-    {'scale': 1.0},
-    {'scale': 3.0},
-    {'scale': 10.0},
-)
+@testing.parameterize(*common_distributions.exponential_params)
 @testing.gpu
 @testing.fix_random()
-class TestExponential(RandomGeneratorTestCase):
-
-    target_method = 'exponential'
-
-    def test_exponential(self):
-        self.generate(scale=self.scale, size=(3, 2))
-
-    @testing.for_dtypes('fd')
-    @_condition.repeat_with_success_at_least(10, 3)
-    def test_exponential_ks(self, dtype):
-        self.check_ks(0.05)(
-            self.scale, size=2000, dtype=dtype)
+class TestExponential(
+    common_distributions.Exponential,
+    RandomGeneratorTestCase
+):
+    pass
 
 
 @testing.parameterize(
@@ -364,34 +257,14 @@ class TestF(RandomGeneratorTestCase):
             self.dfnum, self.dfden, size=2000, dtype=dtype)
 
 
-@testing.parameterize(
-    {'shape': 0.5, 'scale': 0.5},
-    {'shape': 1.0, 'scale': 0.5},
-    {'shape': 3.0, 'scale': 0.5},
-    {'shape': 0.5, 'scale': 1.0},
-    {'shape': 1.0, 'scale': 1.0},
-    {'shape': 3.0, 'scale': 1.0},
-    {'shape': 0.5, 'scale': 3.0},
-    {'shape': 1.0, 'scale': 3.0},
-    {'shape': 3.0, 'scale': 3.0},
-)
+@testing.parameterize(*common_distributions.gamma_params)
 @testing.gpu
 @testing.fix_random()
-class TestGamma(RandomGeneratorTestCase):
-
-    target_method = 'gamma'
-
-    def test_gamma_1(self):
-        self.generate(shape=self.shape, scale=self.scale, size=(3, 2))
-
-    def test_gamma_2(self):
-        self.generate(shape=self.shape, size=(3, 2))
-
-    @testing.for_dtypes('fd')
-    @_condition.repeat_with_success_at_least(10, 3)
-    def test_gamma_ks(self, dtype):
-        self.check_ks(0.05)(
-            self.shape, self.scale, size=2000, dtype=dtype)
+class TestGamma(
+    common_distributions.Gamma,
+    RandomGeneratorTestCase
+):
+    pass
 
 
 @testing.parameterize(
@@ -697,21 +570,14 @@ class TestPareto(RandomGeneratorTestCase):
             a=self.a, size=2000, dtype=dtype)
 
 
-@testing.parameterize(
-    {'lam': 1.0},
-    {'lam': 3.0},
-    {'lam': 10.0},
-)
+@testing.parameterize(*common_distributions.poisson_params)
 @testing.gpu
 @testing.fix_random()
-class TestPoisson(RandomGeneratorTestCase):
-
-    target_method = 'poisson'
-
-    def test_poisson(self):
-        self.generate(lam=self.lam, size=(3, 2))
-
-    # TODO(kataoka): add distribution test
+class TestPoisson(
+    common_distributions.Poisson,
+    RandomGeneratorTestCase
+):
+    pass
 
 
 @testing.parameterize(
@@ -854,25 +720,14 @@ class TestStandardCauchy(RandomGeneratorTestCase):
             size=2000, dtype=dtype)
 
 
-@testing.parameterize(
-    {'shape': 0.5},
-    {'shape': 1.0},
-    {'shape': 3.0},
-)
+@testing.parameterize(*common_distributions.standard_gamma_params)
 @testing.gpu
 @testing.fix_random()
-class TestStandardGamma(RandomGeneratorTestCase):
-
-    target_method = 'standard_gamma'
-
-    def test_standard_gamma(self):
-        self.generate(shape=self.shape, size=(3, 2))
-
-    @testing.for_dtypes('fd')
-    @_condition.repeat_with_success_at_least(10, 3)
-    def test_standard_gamma_ks(self, dtype):
-        self.check_ks(0.05)(
-            shape=self.shape, size=2000, dtype=dtype)
+class TestStandardGamma(
+    common_distributions.StandardGamma,
+    RandomGeneratorTestCase
+):
+    pass
 
 
 @testing.fix_random()
@@ -946,7 +801,7 @@ class TestInterval(RandomGeneratorTestCase):
     @_condition.repeat(3, 10)
     def test_goodness_of_fit(self):
         mx = 5
-        trial = 100
+        trial = 300
         vals = self.generate_many(mx, None, _count=trial)
         vals = [val.get() for val in vals]
         counts = numpy.histogram(vals, bins=numpy.arange(mx + 2))[0]
@@ -1080,6 +935,7 @@ class TestChoiceChi(RandomGeneratorTestCase):
         assert _hypothesis.chi_square_test(counts, expected)
 
     @_condition.repeat(3, 10)
+    @pytest.mark.xfail(runtime.is_hip, reason='ROCm/HIP may have a bug')
     def test_goodness_of_fit_2(self):
         vals = self.generate(3, (5, 20), True, [0.3, 0.3, 0.4]).get()
         counts = numpy.histogram(vals, bins=numpy.arange(4))[0]
@@ -1395,24 +1251,11 @@ class TestSetRandomState(unittest.TestCase):
 
 @testing.gpu
 @testing.fix_random()
-class TestStandardExponential(RandomGeneratorTestCase):
-
-    target_method = 'standard_exponential'
-
-    def test_standard_exponential(self):
-        self.generate(size=(3, 2))
-
-    @_attr.slow
-    @_condition.repeat(10)
-    def test_standard_exponential_isfinite(self):
-        x = self.generate(size=10**7)
-        assert cupy.isfinite(x).all()
-
-    @testing.for_dtypes('fd')
-    @_condition.repeat_with_success_at_least(10, 3)
-    def test_standard_exponential_ks(self, dtype):
-        self.check_ks(0.05)(
-            size=2000, dtype=dtype)
+class TestStandardExponential(
+    common_distributions.StandardExponential,
+    RandomGeneratorTestCase
+):
+    pass
 
 
 @testing.parameterize(
